@@ -89,7 +89,7 @@ typedef struct {
 		CK_ATTRIBUTE_TYPE type;
 		int flags;
 		bool (*validate) (p11_builder *, CK_ATTRIBUTE *);
-	} attrs[32];
+	} attrs[64];
 	CK_ATTRIBUTE * (*populate) (p11_builder *, p11_index *, CK_ATTRIBUTE *);
 	CK_RV (*validate) (p11_builder *, CK_ATTRIBUTE *, CK_ATTRIBUTE *);
 } builder_schema;
@@ -992,14 +992,23 @@ const static builder_schema trust_schema = {
 	  { CKA_ISSUER, CREATE },
 	  { CKA_SUBJECT, CREATE },
 	  { CKA_SERIAL_NUMBER, CREATE },
+	  /* official trust attributes */
 	  { CKA_TRUST_SERVER_AUTH, CREATE },
 	  { CKA_TRUST_CLIENT_AUTH, CREATE },
-	  { CKA_TRUST_EMAIL_PROTECTION, CREATE },
 	  { CKA_TRUST_CODE_SIGNING, CREATE },
+	  { CKA_TRUST_EMAIL_PROTECTION, CREATE },
+	  { CKA_TRUST_IPSEC_IKE, CREATE },
+	  { CKA_TRUST_TIME_STAMPING, CREATE },
+	  { CKA_TRUST_OCSP_SIGNING, CREATE },
+	  /* vendor trust attributes previuosly used by NSS */
+	  { CKA_NSS_TRUST_SERVER_AUTH, CREATE },
+	  { CKA_NSS_TRUST_CLIENT_AUTH, CREATE },
+	  { CKA_NSS_TRUST_EMAIL_PROTECTION, CREATE },
+	  { CKA_NSS_TRUST_CODE_SIGNING, CREATE },
 	  { CKA_NSS_TRUST_IPSEC_END_SYSTEM, CREATE },
 	  { CKA_NSS_TRUST_IPSEC_TUNNEL, CREATE },
 	  { CKA_NSS_TRUST_IPSEC_USER, CREATE },
-	  { CKA_TRUST_TIME_STAMPING, CREATE },
+	  { CKA_NSS_TRUST_TIME_STAMPING, CREATE },
 	  { CKA_NSS_TRUST_DIGITAL_SIGNATURE, CREATE },
 	  { CKA_NSS_TRUST_NON_REPUDIATION, CREATE },
 	  { CKA_NSS_TRUST_KEY_ENCIPHERMENT, CREATE },
@@ -1354,14 +1363,22 @@ build_trust_object_eku (CK_ATTRIBUTE *object,
 		CK_ATTRIBUTE_TYPE type;
 		const char *oid;
 	} eku_attribute_map[] = {
+		/* official trust attributes */
 		{ CKA_TRUST_SERVER_AUTH, P11_OID_SERVER_AUTH_STR },
 		{ CKA_TRUST_CLIENT_AUTH, P11_OID_CLIENT_AUTH_STR },
 		{ CKA_TRUST_CODE_SIGNING, P11_OID_CODE_SIGNING_STR },
 		{ CKA_TRUST_EMAIL_PROTECTION, P11_OID_EMAIL_PROTECTION_STR },
+		{ CKA_TRUST_TIME_STAMPING, P11_OID_TIME_STAMPING_STR },
+		/* vendor trust attributes previuosly used by NSS */
+		{ CKA_NSS_TRUST_SERVER_AUTH, P11_OID_SERVER_AUTH_STR },
+		{ CKA_NSS_TRUST_CLIENT_AUTH, P11_OID_CLIENT_AUTH_STR },
+		{ CKA_NSS_TRUST_CODE_SIGNING, P11_OID_CODE_SIGNING_STR },
+		{ CKA_NSS_TRUST_EMAIL_PROTECTION, P11_OID_EMAIL_PROTECTION_STR },
+		{ CKA_NSS_TRUST_TIME_STAMPING, P11_OID_TIME_STAMPING_STR },
 		{ CKA_NSS_TRUST_IPSEC_END_SYSTEM, P11_OID_IPSEC_END_SYSTEM_STR },
 		{ CKA_NSS_TRUST_IPSEC_TUNNEL, P11_OID_IPSEC_TUNNEL_STR },
 		{ CKA_NSS_TRUST_IPSEC_USER, P11_OID_IPSEC_USER_STR },
-		{ CKA_TRUST_TIME_STAMPING, P11_OID_TIME_STAMPING_STR },
+		{ CKA_NSS_TRUST_TIME_STAMPING, P11_OID_TIME_STAMPING_STR },
 		{ CKA_INVALID },
 	};
 
